# -*- coding: utf-8 -*-
"""Types used in the packages."""
from typing import Literal, Optional
from pandas import Timestamp

bucketT = Literal['1m', '5m', '1h', '1d']
oTimestampT = Optional[Timestamp]
symbolT = Literal['XBTUSD', 'ADAM20']



# -*- coding: utf-8 -*-
"""Settings.py."""
LIVE_URL = "https://www.bitmex.com/api/v1/"
TEST_URL = "https://testnet.bitmex.com/api/v1/"

# Duration converter
TC = {"1m": "60s", "5m": "300s", "1h": "1H", "1d": "1D"}

# the oldest date I know off but 2015-09-26 for bitmex
# Should be in UTC
STARTDATE_DFT = {"XBTUSD": "2016-05-05 04:00", "ADAM20": "2020-03-14"}



# import pandas as pd
import os
import os.path as op
import re
import datetime as dt
import logging

from pandas import DataFrame, to_datetime
from pandas.errors import OutOfBoundsDatetime

""" Bibliothèques d'utilitaire pour manipuler les fichiers """


def is_file(fname):
    """Renvois true or false depending if file pas condition."""
    #  isabs isdir isfile islink ismount
    return op.isfile(fname)


def fcond(f, pat):
    """Set a condition on a file name."""
    rpat = re.compile(pat)
    return rpat.search(f)


def is_xls_file(f):
    return fcond(f, pat=r"\.xls$")


def is_csv_file(f):
    return fcond(f, pat=r"\.csv$")


def sort_param(fname):
    """renvois un nombre qui permet de classer le fichier fname"""
    #  getatime getctime *getmtime* getsize
    return op.getmtime(fname)


def get_latest_export_basename(srcdir=None):
    """returns the latest export file name.  If the srcdir is not given returns the cwd"""
    "FOO_Baz_export_dossiers_heure_minute_year_month_day.xls"
    fname = get_cond_recent_file(rang=0, srcdir=srcdir, coln="fname", fcond=is_file)
    bname = "".join(fname.split(".")[:-1])
    return bname


def get_recent_file(rang=0, srcdir=None):
    """soit un rang 0 à nb fichiers dans srcdir, renvois la date du fichier à ce rang class du plus récent au plus ancien.
   srcdir (=cwd by default)"""
    return get_cond_recent_file(rang=rang, srcdir=srcdir, coln="fname", fcond=is_file)


def get_recent_xls_file(rang=0, srcdir=None, dcond=None):
    return get_cond_recent_file(
        rang=rang, srcdir=srcdir, coln="fname", fcond=is_xls_file, dcond=dcond
    )


def get_most_recent_csv_file(srcdir=None):
    return get_cond_recent_file(srcdir=srcdir, rang=0, coln="fname", fcond=is_csv_file)


def get_recent_date(rang=0, srcdir=None):
    """soit un rang 0 à nb fichiers dans srcdir, renvois la date du fichier à ce rang class du plus récent au plus ancien.
   srcdir (=cwd by default)"""
    return get_cond_recent_file(
        srcdir=srcdir, rang=rang, coln="getmtime", fcond=is_xls_file
    )


def get_cond_recent_file(
    rang=0, srcdir=None, coln="fname", fcond=lambda x: op.isfile(x), dcond=None
):
    """soit un rang 0 à nb fichiers dans srcdir, renvois la date du fichier à ce rang class du plus récent au plus ancien si pat (def \.xls$) en regex est dans sont fullname (avec path)
    srcdir (=cwd by default)"""
    return get_cond_recent_info(
        rang=rang, srcdir=srcdir, coln=coln, fcond=fcond, dcond=dcond
    )


def get_cond_recent_info(
    rang=0, srcdir=None, coln="fname", fcond=lambda x: op.isfile(x), dcond=None
):
    """soit un rang 0 à nb fichiers dans srcdir, renvois la date du fichier à ce rang class du plus récent au plus ancien si pat (def \.xls$) en regex est dans sont fullname (avec path)
    srcdir (=cwd by default)
    dcond: condition sur la date. une fonction du type lambda d: d < ma date"""
    if srcdir is None:
        srcdir = os.getcwd()

    # get list of all files in dir
    fullnames = ["/".join([srcdir, f]) for f in os.listdir(srcdir)]

    # keep file (type f) and adjoinct a getmtime value
    fnames = [(f, op.getmtime(f)) for f in fullnames if fcond(f)]
    # print(fnames)
    # put this in a dataFrame
    colfilen = "fname"
    coldaten = "getmtime"
    df = DataFrame(fnames, columns=[colfilen, coldaten])

    # parse date col to_datetime format
    df.loc[:, coldaten] = to_datetime(df.loc[:, coldaten], unit="s")
    # filtering dates
    if dcond is not None:
        mask = df.loc[:, coldaten].apply(dcond)
        df = df.loc[mask, :]

    # sort
    df = df.sort_values(coldaten)[::-1]

    res = None

    try:
        if coln == "all":
            res = df.iloc[rang].loc[:]
        else:
            res = df.iloc[rang].loc[coln]

    except IndexError:
        logging.warning(f"srcdir={srcdir} et fcond={fcond}")
        logging.warning(f"rang={rang} et coln={coln} dans df(head)\n%s" % df.head(3))
        raise

    return res


def remove_ieme_xls_file(ieme=0, srcdir=None):
    """Supprime la ieme xls file sorted by name, if ieme = * les supprime toutes"""
    return remove_ieme_cond_ordered_files(
        ieme=ieme, srcdir=srcdir, cond=("extension", ext_is_xls)
    )


def ext_is_xls(ext):
    """takes an extention and say if it is xls"""
    return ext == ".xls"


def remove_ieme_cond_ordered_files(ieme, srcdir=None, cond=None, order=None):
    """cols can be: fullname, mtime, atime, ctime, size, isdir, isfile, islink, ext, path, bname. sorte by fullname by default, renvois les fichiers supprimés"""
    dff = get_df_files(srcdir)

    if order is None:
        sort_by, ascending = "fullname", True
    else:
        sort_by, ascending = order

    if cond is None:
        ffnames = dff.sort_values(by=sort_by, ascending=ascending).fullname
    else:
        ffnames = (
            dff.where(dff.loc[:, cond[0]].apply(cond[1]))
            .dropna()
            .sort_values(by=sort_by, ascending=ascending)
            .fullname
        )

    if ieme == "*":
        rfiles = []
        for ffname in ffnames:
            os.remove(ffname)
            rfiles.append(ffname)
            pass
        return rfiles
    else:
        os.remove(ffnames[ieme])
        return [ffnames[ieme]]


def remove_but_last_csv(srcdir=None):
    """remove all but most recent csv file"""
    if srcdir is None:
        cwd = os.getcwd()
        raise Exception(f"Vous devez fournir un srcdir. cwd={cwd}")

    dff = get_df_files(srcdir)
    ffnames = (
        dff.where(dff.loc[:, "extension"].apply(lambda ext: ext != ".csv"))
        .dropna()
        .fullname
    )
    rfiles = []
    for ffname in ffnames:
        os.remove(ffname)
        rfiles.append(ffname)

    ffnames = get_df_files(srcdir).sort_values(by="mtime").fullname

    if len(ffnames) > 1:
        for ffname in ffnames[:-1]:
            os.remove(ffname)
            rfiles.append(ffname)

    return rfiles


def remove_but_last_5(srcdir=None):
    """remove all but most recent csv file"""
    if srcdir is None:
        cwd = os.getcwd()
        raise Exception(f"Vous devez fournir un srcdir. cwd={cwd}")

    # dff = get_df_files(srcdir)
    ffnames = get_df_files(srcdir).sort_values(by="mtime").fullname

    rfiles = list()
    n = len(ffnames)
    if n > 5:
        for ffname in ffnames[: n - 5]:
            os.remove(ffname)
            rfiles.append(ffname)

    return rfiles


def get_df_files(srcdir=None):
    """Renvois un df avec un max d'info sur les files du srcdir
cols can be: fullname, mtime, atime, ctime, size, isdir, isfile, islink, ext, path, bname"""

    if srcdir is None:
        srcdir = os.getcwd()

    fullname = ["/".join([srcdir, f]) for f in os.listdir(srcdir)]
    df = DataFrame(data=fullname, columns=["fullname"])

    INFOS = {
        "mtime": op.getmtime,
        "atime": op.getatime,
        "ctime": op.getctime,
        "size": op.getsize,
        "isdir": op.isdir,
        "isfile": op.isfile,
        "islink": op.islink,
        "extension": lambda f: op.splitext(f)[1],
        "path": lambda f: op.split(f)[0],
        "bname": lambda f: op.split(f)[1],
    }

    for col, func in INFOS.items():
        df.loc[:, col] = df.fullname.apply(func)
        if "time" in col:
            df.loc[:, col] = to_datetime(df.fullname.apply(func), unit="s")
            df.loc[:, f"r{col}"] = df.loc[:, col].apply(lambda d: d.round("1s"))

    return df


def get_extraction_date(ffname, offset=60):
    """
    Parse the ffname which should be an extraction file to get the date from it.

    return the date - offset (default 3600s)
    """
    # extracting the date from the name
    pat = r"_(?P<day>\d+)_(?P<month>\d+)_(?P<year>\d+)_(?P<hour>\d+)_(?P<minute>\d+)\..{3}$"
    # print(f'pat={pat}\n{ffname}')
    trouvailles = re.compile(pat).search(ffname)
    if trouvailles is None:
        # au cas ou le nom du fichier n'est pas bien formé.  on prend la date du jour.
        dpextract = dt.datetime.now()
        pass
    else:
        trouvailles = trouvailles.groupdict()
        # convertir en int et crée la date
        # to get date previous extraction (qui si généré par le système doit être en phase avec le contenu de la df)
        dpextract = dt.datetime(**{k: int(v) for (k, v) in trouvailles.items()})

    # On lui enlève une heure créer un chevauchement et ne pas rater de mise à jour
    # sachant qu'il y a peu de nouveaux dossiers dans l'heure qui suit les extractions
    return dpextract - dt.timedelta(seconds=offset)


def get_previous_extraction_date(srcdir="./Data"):
    """get the second to the most recent extraction file name from dir srcdir,
    parse the name to get the date from it and return the date"""
    # before most recent file name
    bnames = get_df_files(srcdir).bname.sort_values()
    n = len(bnames)
    if n > 1:
        bmrfn = bnames.iloc[-2]
    elif n == 1:
        bmrfn = bnames.iloc[0]
    else:
        return False

    return get_extraction_date(bmrfn)


def une_date_est_plus_jeune(serieDeDates, dateref):
    """compare les dates d'une série avec une date de référence et renvois vrai si l'une des dates de la série est plus jeunes (plus grande) que la date de référence"""
    # eventuellement si serieDeDate is pd.serie
    # (serieDeDates < dateref).any()
    return any([dateref < d for d in serieDeDates])


def can_be_date(dt):
    return not (dt in ["", None] or pd.isna(dt))


def datetime_conversion(df, datetimecols):
    for coln in datetimecols:
        df.loc[:, coln] = df.loc[:, coln].apply(datetime_conv)
    return df


def simpledate_conversion(df, puredatecols):
    for coln in puredatecols:
        df.loc[:, coln] = df.loc[:, coln].apply(date_conv)
    return df


# #### Date conversion ####
def date_conv(dn, gedci_date=False, day_first=True):
    """Soit dn une date au format string, vérifie des erreurs et renvois la date au format %Y-%d-%m
    -dans le cas de gedci_date=True, on a des dates avec day first et moins de condition sur l'année"""
    if gedci_date:
        pat = r"(?P<day>\d{2})(?:_)(?P<month>\d{2})(?:_)(?P<year>\d{4})(?:_)(?P<hour>\d{2})(?:_)(?P<minute>\d{2})"
    else:
        pat = r"(?P<year>\d{4})(?:-|/)(?P<month>\d{2})(?:/|-)(?P<day>\d{2})"
        pat2 = r"(?P<day>\d{2})(?:-|/)(?P<month>\d{2})(?:/|-)(?P<year>\d{4})"

    if can_be_date(dn):
        try:
            try:
                trouvailles = re.compile(pat).search(dn)
                if trouvailles is None:
                    trouvailles = re.compile(pat2).search(dn)  # could be none again
                trvD = trouvailles.groupdict()
                day, month, year = map(int, [trvD["day"], trvD["month"], trvD["year"]])

                if not day_first:
                    day, month = month, day

                if gedci_date:
                    hour, minute = map(int, trvD["hour"], trvD["minute"])
                    year = 1900 if (1930 > year) else year
                    hour = 0 if hour > 24 else hour
                    minute = 0 if minute > 60 else minute
                else:
                    # pour une date de naissance ok
                    if (1930 > year) or (year > (pd.Timestamp.now().year - 10)):
                        year = 1900

                    # mauvaise partique ce sont des erreurs ignore le format américain des dates
                    month = 1 if month > 12 else month
                    daysinmonth = pd.Timestamp(
                        year=year, month=month, day=1
                    ).daysinmonth
                    day = 1 if day > daysinmonth else day

                try:
                    if gedci_date:
                        return dt.datetime(
                            year=year, month=month, day=day, hour=hour, minute=minute
                        )
                    else:
                        return pd.Timestamp(year=year, month=month, day=day).strftime(
                            "%Y-%m-%d"
                        )

                except OutOfBoundsDatetime as obd:
                    logging.error(f"pb avec {dn} dans {obd}")
                    raise obd

            except AttributeError as ae:
                logging.debug(
                    f"Error {ae}:\n {dn}=dn, pat={pat}, can_be_date={can_be_date(dn)}, gedci_date={gedci_date}"
                )
                pass

        except Exception as e:
            logging.warning(f"Check {dn}")
            raise (e)

    return None


def datetime_conv(dt):
    """Soit une Datetime dt la converti en Y-m-d"""
    if can_be_date(dt):
        try:
            ts = pd.to_datetime(dt, dayfirst=True)
            return ts.strftime("%Y-%m-%d %H:%M")
        except ValueError:
            logging.warning("dt of type %s = '%s'" % (type(dt), dt))
            raise
    return None


def dt_equal(dt1, dt2, resolution=1):
    """compare des date selon une resolution en secondes (default 1)
    peut recevoir des dates en string ou de divers format.  le convertis en avant comparaison
    """
    # tt1, tt2 = type(dt1), type(dt2)
    # logging.debug(f"type(dt1)={tt1}, type(dt2)={tt2}, dt1={dt1}, dt2={dt2}")
    if dt1 is None and dt2 is None:
        return True

    if can_be_date(dt1) and can_be_date(dt2):
        try:
            ts1 = pd.Timestamp(dt1).round("ms")
            ts2 = pd.Timestamp(dt2).round("ms")
            return abs(ts1 - ts2).total_seconds() < resolution
        except Exception as e:
            logging.info(f"ts1={ts1}, ts2={ts2} et dif {ts1 - ts2}")
            raise (e)

    return None



# -*- coding: utf-8 -*-
"""main file getBitMEXData.py."""
from time import sleep
from typing import Tuple

import argparse
import logging
import os
import platform  # handle os check
import time
import sys
import requests as rq
from pandas import DataFrame, Timestamp, Timedelta
from pathlib import Path

from getBitMEXData.getBitMEXtypes import bucketT, oTimestampT, symbolT
from getBitMEXData.settings import LIVE_URL, TEST_URL, TC, STARTDATE_DFT

# setting the default time zone for the system
if platform.system() == "Linux":
    # Time.tzset ne fonctionne qu'avec UNIX le mettre en commentaire pour windows
    # mais veillé à spécifier la date de départ et de fin des messages.  avec les option --startTime et --endTime
    OS_TZ = os.environ.get("TZ", "UTC")
else:
    # in the case os.environ does not exist
    OZ_TZ = "UTC"

STRF = "%Y-%m-%d__%H_%M"  # default time format for saving the data

logger = logging.getLogger()
logger.setLevel("INFO")

# Converts bitmex time unit to pd.timestamp time units
URLS = {
    True: LIVE_URL,
    False: TEST_URL,
}


def init_session(name="foo"):
    """Set a session with name `foo`."""
    sess = rq.Session()
    sess.headers.update({"user-agent": f"{name}-"})
    sess.headers.update({"content-type": "application/json"})
    sess.headers.update({"accept": "application/json"})
    return sess


def make_request(query, sess, auth, url, verb="GET"):
    """
    Make the request with query been passed via rest in sessions sess.

    - request verb  (default GET)
    """
    rep, req = None, None
    try:
        req = rq.Request(verb, f"{url}", auth=auth, params=query)
        print(req.url, query)
        prepp = sess.prepare_request(req)
        rep = sess.send(prepp, timeout=10)
        rep = rep.json()
        logmsg = (
            f"Req: {req}, url={req.url}, parms={query}, auth={auth}"
            f"Prepp: {prepp}: body={prepp.body}, header={prepp.headers},"
            f" url={prepp.path_url}"
        )
        logger.debug(logmsg)
    except Exception as e:
        if req:
            logmsg = (
                f"#### ERROR ####\nReq: req={req}, url={req.url}, params={query}"
                f"\nPrepp: header={prepp.headers}, url={prepp.path_url}"
            )
            logmsg += f"\nRep: {rep}"

            logger.exception(logmsg)
        raise e

    return rep


def get_time_window(rep, reverse=False) -> Tuple[Timestamp, Timestamp]:
    """
    Return first and last date of a request response.

    - reverse if False (default), oldest date is first,
    """
    try:
        old = Timestamp(rep[0]["timestamp"])
        recent = Timestamp(rep[-1]["timestamp"])
    except (KeyError, IndexError) as e:
        logger.error(f"rep={rep}, probably empty.")
        raise e

    return (recent, old) if reverse else (old, recent)


def get_bucketed_trades(
    url,
    Q=None,
    fout=None,
    startTime: oTimestampT = None,
    endTime: oTimestampT = None,
    binSize: bucketT = "1d",
    pause: float = 0.5,
    reverse: str = "false",
    symbol: symbolT = "XBTUSD",
):
    """
    Returns the historical data from bitMEX (default).

    resulting values are : timestamp, symbol, open, high, low, close, trades,
    volume,  vwap,  lastSize, turnover, homeNotional, foreignNotional,

    - url : live or test url
    Times are in isoformat eg. 2016-12-27T11:00Z
    - `startTime`: date of first record to download
    - `endTime` : date of the last recorde to download
    - binSize : bucketed size to ask: one of 1m, 5m, 1h, 1d

    Params:
    - Q : The Query requested.
    should be a dictionnary with keys binSize, partial, symbol, count and reverse.
    - fout : the name of the file to write to the results to
    - pause : to throttle the requests and avoid been rejected by bitMEX
    - reverse : should we return earliest data first ?
    """
    logger.debug(
        f"Got {Q}, fout={fout}, startTime={startTime}, endTime={endTime},"
        f" binSize={binSize}, pause={pause}"
    )

    # Init session and defaults settings
    auth = None  # auth = APIKeyAuthWithExpires(apiKey, apiSecret)
    sess = init_session()
    fout = "./btxData-{binSize}-{endTime.strftime(STRF)}.csv" if fout is None else fout
    # prise en compte de windows
    fout = Path(fout)

    Q = (
        {
            "binSize": binSize,
            "partial": "false",
            "symbol": symbol,
            "count": 180,
            "reverse": reverse,
        }
        if Q is None
        else Q
    )

    if startTime is not None:
        Q["startTime"] = Timestamp(startTime).round(TC[binSize])
    else:
        startTime = Q["startTime"]

    # Ready to open the file make requests and write results
    with open(fout, "w") as fd:
        Q, firstReqDate, lastReqDate = request_write_nlog(
            Q, sess, auth, url, fd, header=True, pause=0
        )
        logging.debug(f"Req 0: Q={Q})")
        logging.info(f"{firstReqDate.strftime(STRF)}:{lastReqDate.strftime(STRF)}")

        i = 1
        while not reached(lastReqDate, endTime):
            Q, firstReqDate, lastReqDate = request_write_nlog(
                Q, sess, auth, url, fd, step=i, startTime=lastReqDate, pause=pause
            )
            i += 1

    # last log before exit
    Q["startTime"], Q["endTime"] = startTime, endTime
    logging.warning(f"Finished in {i} requests for query={Q}")

    return sess


def request_write_nlog(
    query, sess, auth, url, fd, header=False, pause=1, step=0, startTime=None
):
    """
    Make the requests and write the results in a file descriptor.

    returns the query and 2 timestamps.
    """
    logger.debug(f"Requesting {query}")
    # pause to avoid been rejected, below 1.2 s between requests,
    # it can be rejected by server after a few tens of requests.
    sleep(pause)

    if startTime is not None:
        query["startTime"] = startTime
    try:
        rep = make_request(query, sess, auth, url)
        firstReqDate, lastReqDate = get_time_window(rep)
    except (KeyError, IndexError):
        # Then we probably have no data for this request, we jump to next day
        firstReqDate = Timestamp(query.get("startTime"))
        lastReqDate = Timestamp(firstReqDate) + Timedelta("1D")
        logmsg = (
            f"# Empty Response: Step={step}, {firstReqDate.strftime(STRF)}"
            f" --> {lastReqDate.strftime(STRF)}.\n"
        )
    except Exception:
        logger.error(f"query={query}, auth={auth}, url={url}")
        raise

    else:
        df = DataFrame(rep).set_index("timestamp")
        df.to_csv(fd, header=header)
        logmsg = (
            f"# Step={step}, {firstReqDate.strftime(STRF)}"
            f" --> {lastReqDate.strftime(STRF)}.\n"
        )

    fd.write(logmsg)
    print(logmsg, end="\r")

    return query, firstReqDate, lastReqDate


def reached(lastReqDate, endTime=None):
    """
    Returns True si endTime <= lastReqDate (end more recent than last)
    or if endTime is None
    """
    endTz = OS_TZ if endTime is None or endTime.tz is None else endTime.tz
    lastTz = OS_TZ if lastReqDate.tz is None else lastReqDate.tz

    return endTime is None or Timestamp(endTime).tz_convert(endTz) <= Timestamp(
        lastReqDate
    ).tz_convert(lastTz)


def parse_args():
    """Settings the applications's arguments and options."""
    description = """An application to download bitmex's data with what ever resolution you need."""
    fout_dft = "btx"
    fout_help = f"base Name of the csv file where to save the results. (default {fout_dft}<symbol>-<BINSIZE>-<LASTRECORDDATE>.csv)"
    count_dft = 600
    count_help = "Max number each of records in requests (default 600)"
    pause_dft = 1.2
    pause_help = "Min time to wait between 2 requests (default 1.2).  to avoid overloading the server"
    binSize_dft = "1d"
    binSize_help = "Bin size or type requested, or time resolution (default 1d), can also be 1m, 5m, 1h."
    startTime_dft = None
    startTime_help = "Time to start the data collection (default, oldest available 2016-05-05 04:00:00 'UTC').  Check time zones"
    endTime_dft = None
    endTime_help = "Time to end the data collection (default, now - 1 unit of chosen resolution)-05-05 04:00:00 'UTC').  Check TZ"
    logLevel_dft = "WARNING"
    logLevel_help = "set the log level"
    live_help = "If present use LIVE keys to get the data else use the test site."
    entryPoint_dft = "trade/bucketed"
    entryPoint_help = "Set the entry level.  the path to append to the LIVE or TEST url before the query"
    symbol_help = "Set the symbol for which to get historical data def. XBTUSD.  default start date may change depending on symbol"
    symbol_dft = "XBTUSD"

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("--fout", "-f", help=fout_help, default=fout_dft)
    parser.add_argument("--count", "-c", type=int, help=count_help, default=count_dft)
    parser.add_argument("--pause", "-p", type=float, help=pause_help, default=pause_dft)
    parser.add_argument("--binSize", "-b", help=binSize_help, default=binSize_dft)
    parser.add_argument("--startTime", "-s", help=startTime_help, default=startTime_dft)
    parser.add_argument("--endTime", "-e", help=endTime_help, default=endTime_dft)
    parser.add_argument("--live", "-l", action="store_true", help=live_help)
    parser.add_argument("--logLevel", "-L", help=logLevel_help, default=logLevel_dft)
    parser.add_argument(
        "--entryPoint", "-E", help=entryPoint_help, default=entryPoint_dft
    )
    parser.add_argument("--symbol", "-S", help=symbol_help, default=symbol_dft)

    return parser.parse_args()


def main_prg():
    """Run the main programme."""
    args = parse_args()

    logger.setLevel(args.logLevel)

    timeUnit = TC[args.binSize]
    nUnit, tUnit = int(timeUnit[:-1]), timeUnit[-1]

    startTime = (
        Timestamp(STARTDATE_DFT[args.symbol]).round(timeUnit)
        if args.startTime is None
        else Timestamp(args.startTime)
    )
    # localising the timezone
    startTime = startTime.tz_localize(OS_TZ)

    # To avoid empty request we stop one unit befor the present date.
    endTime = (
        (Timestamp.now() - Timedelta(nUnit, tUnit)).round(timeUnit)
        if args.endTime is None
        else Timestamp(args.endTime)
    )

    # making_sure the format is also valid for windows
    endTime = endTime.tz_localize(OS_TZ)  # .strftime(STRF)

    query = {
        "binSize": args.binSize,
        "count": args.count,
        "partial": "false",
        "reverse": "false",
        "symbol": args.symbol,
    }

    # kwargs stand for key words arguments
    kwargs = {
        "endTime": endTime,
        "fout": f"{args.fout}{args.symbol}-{args.binSize}-{endTime}",
        "pause": args.pause,
        "startTime": startTime,
    }

    # use live or test ids
    URL = URLS[args.live]

    logger.warning(f"Writting data to {kwargs['fout']}")

    _ = get_bucketed_trades(url=f"{URL}{args.entryPoint}", Q=query, **kwargs)
    return None


if __name__ == "__main__":
    main_prg()
    sys.exit()



# -*- coding: utf-8 -*-
""" Fichier avec des fonctions pour récupérer les données historique de BitMex en ligne """
import argparse
import subprocess as sp
import logging
import os
import time
import sys

from pandas import Timestamp, Timedelta

import getBitMEXData.utils as u
from getBitMEXData.settings import LIVE_URL, TEST_URL, TC
from getBitMEXData.getBitMEXData import get_bucketed_trades


os.environ["TZ"] = "UTC"
time.tzset()

# {"1m": "60s", "5m": "300s", "1h": "1H", "1d": "1D"}

logger = logging.getLogger()
STRF = "%Y-%m-%d %H:%M"  # default time format

# Converts bitmex time unit to timestamp time units
URLS = {
    True: (LIVE_URL),
    False: (TEST_URL),
}


def get_last_record(fullname: str) -> str:
    """Renvois le dernier enregistrement d'un fichier btxData with fullname."""
    # on récupère les dernières lignes du fichier
    _lrs = sp.run(f"tail {fullname}".split(), stdout=sp.PIPE).stdout.split(b"\n")

    # on les filtres
    lastRecords = [lr.decode() for lr in _lrs if not lr.startswith(b"#") and len(lr)]

    assert len(
        lastRecords
    ), f"Vérifier qu'il y a bien des enregistrements valides dans fullname={fullname}"

    return lastRecords[-1]


def get_record_date(record) -> Timestamp:
    """
    Get the Timestamp from the record row.

    Keyword Arguments:
    record -- un enregistrement valid de données historique de bitmex binned en 1m
    """
    return Timestamp(record.split(",")[0])


def get_recent_record_file_date(fullname: str) -> Timestamp:
    """
    Get the most recent record date from fullname.
    
    get it from the last record of the file.
    Keyword Arguments:
    fullname -- nom du fichier
    """
    return get_record_date(get_last_record(fullname))


def get_recent_data(
    fromdate: Timestamp, binsize: str = "1m", fout: str = "tmp.csv", live: bool = False
):
    """
    Load Bitmex's data fromdate in binsize and save them in fout.

    Keyword Arguments:
    fromdate -- date à laquelle commencer le téléchargement
    fout     --  fichier de sortie
    live  -- use live or test keys
    """
    # {"1m": "60s", "5m": "300s", "1h": "1H", "1d": "1D"}
    binSize = binsize
    timeUnit = TC[binSize]
    _, tUnit = int(timeUnit[:-1]), timeUnit[-1]

    startTime = fromdate
    endTime = (Timestamp.now() - Timedelta(2, tUnit)).round(timeUnit)

    query = {
        "binSize": binSize,
        "count": 600,
        "partial": "false",
        "reverse": "false",
        # need to get the symbol from records too or from file name
        "symbol": "XBTUSD",
    }
    kwargs = {"endTime": endTime, "fout": fout, "pause": 1.2, "startTime": startTime}

    # use live or test ids
    URL = URLS[live]
    URL = URL + "trade/bucketed"
    logger.warning(f"{URL}, {query}, {kwargs}")
    sess = get_bucketed_trades(URL, Q=query, **kwargs)
    return sess


def update_file(fname: str, fout: str = "./tmp.csv", live: bool = False):
    """
    Write the fname updated with new data.

    Keyword Arguments:
    fname -- name of the file to update. should be name btxData-{freq}-{lastudpatedate}.csv

    Get the freq (ie bin size) from the file name.
    Download new data in tmp.csv
    Append tmp.csv to original file but update the name with last updatedate
    """
    # get
    oldDate = get_recent_record_file_date(fname)
    binSize = get_fname_binsize(fname)
    # download data in fout
    get_recent_data(oldDate, binSize, fout, live)
    concat_files(fname, fout, removetmp=True)
    return update_name(fname)


def update_name(fname: str):
    """
    Keyword Arguments:
    fname -- name of the file to update

    Returns: updates the name of the file with date of the last record in it
    "fout": f"{args.fout}-{args.binSize}-{endTime}.csv"""
    newDate = get_recent_record_file_date(fname).strftime("%Y%m%dT%H:%M")
    simpDate = get_recent_record_file_date(fname).strftime("%Y")
    binSize = get_fname_binsize(fname)
    baseName = get_fname_basename(fname)
    newName = f"{baseName}-{binSize}-{newDate}.csv"
    simpName = f"{baseName}-{binSize}-{simpDate}.csv"
    logger.warning(f"Renaming {fname} -> {newName}\n" f"Updating link {simpName}")
    os.rename(fname, newName)
    try:
        os.remove(simpName)
    except FileNotFoundError:
        pass

    os.symlink(newName, simpName)

    return


def get_fname_binsize(fname: str) -> str:
    """If the name is {base}-{binsize}-{date}.csv return the binsize."""
    return fname.split("-")[1]


def get_fname_basename(fname: str) -> str:
    """If the name is {base}-{binsize}-{date}.csv return the basename."""
    return fname.split("-")[0]


def concat_files(file1: str, file2: str, removetmp: bool = False):
    """
    Write a file fname which is the concatenation of file1 and file2.

    Keyword Arguments:
    file1 -- first file (to be on the top)
    file2 -- second or tail fail
    Files should have the same format
    concat file 2 at the end of file 1
    removeoriginal (False) -- should we remove file1 and 2 after concatenation
    """
    # pb this erased data
    my_cmd = ["tail", "-n", "+2", file2]
    with open(file1, "a") as f:
        _ = sp.run(my_cmd, stdout=f)
    if removetmp:
        logger.warning(f"Removing {file2}.")
        return sp.run(f"rm {file2}".split(), stdout=sp.PIPE)
    else:
        return _


def main(fname: str, live: bool):
    """Update the btxData file fname with latest data."""
    update_file(fname, live=live)


def parse_args():
    """Parse arguments."""
    parser = argparse.ArgumentParser(
        description="Utility to update btxDatafile with latest data"
    )

    fname_help = (
        f"name of the file to update.  if None takes the latest"
        " btxDatafile in current directory"
    )
    live_help = f"If present use live historic data"
    logLevel_def = "INFO"
    logLevel_help = f"Set logLevel (default {logLevel_def})"
    parser.add_argument("--fname", "-f", help=fname_help, default=None)
    parser.add_argument("--live", "-l", action="store_true", help=live_help)
    parser.add_argument("--logLevel", "-L", help=logLevel_help, default=logLevel_def)

    return parser.parse_args()


def main_prg():
    """Main program."""
    args = parse_args()
    logger = logging.getLogger(args.logLevel)

    if args.fname is None:

        def fcond(f):
            return u.fcond(f, pat=r"btxData.*csv$")

        fname = u.get_cond_recent_file(fcond=fcond)
    else:
        fname = args.fname

    logger.warning(f"Running {'Live' if args.live else 'Test'} with {fname}")
    main(fname, args.live)


if __name__ == "__main__":
    main_prg()
    sys.exit()



# -*- coding: utf-8 -*-
"""
Setup.py.

Check https://packaging.python.org/tutorials/packaging-projects/
and file in python/Docs/python-in-nutshell.pdf
upload to https://test.pypi.org/manage/projects/.
https://setuptools.readthedocs.io/en/latest/setuptools.html#including-data-files
"""
from setuptools import setup, find_packages

with open("README.rst", "r") as f:
    long_description = f.read()

setup(
    name="getBitMEXData",
    version="0.1.3",
    description="Utility to download bitMEX historical price data",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="Malik Koné",
    author_email="malikykone@gmail.com",
    url="https://github.com/maliky/getBitmexData",
    packages=find_packages(),
    zip_safe=False,
    python_requires=">=3.8",
    entry_points={
        "console_scripts": [
            "getBitMEXData=getBitMEXData.getBitMEXData:main_prg",
            "update-kola=getBitMEXData.update_kola:main_prg",            
        ]
    },
    install_requires=[
        "pandas",
        # "numpy",
        # "websocket_client",
        "requests",
    ],
    extras_require={
        'dev': ['mypy', 'flake8', 'black'],
        'packaging': ['twine'],
        "test": ['pytest', 'hypothesis'],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Natural Language :: French",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Topic :: Office/Business :: Financial",
        "Topic :: Utilities",
    ],
    package_data={"Doc": ["*txt"]},
)



# -*- coding: utf-8 -*-
from requests.auth import AuthBase
from urllib.parse import urlparse

import hashlib
import hmac
import time

# ###  Regroupe les fonctions et classe liées  à l'authentification####


class AccessTokenAuth(AuthBase):

    """Attaches Access Token Authentication to the given Request object."""

    def __init__(self, accessToken):
        """Init with Token."""
        self.token = accessToken

    def __call__(self, r):
        """Called when forming a request - generates access token header."""
        if self.token:
            r.headers["access-token"] = self.token
        return r


class APIKeyAuthWithExpires(AuthBase):

    """Attaches API Key Authentication to the given Request object. This implementation uses `expires`."""

    def __init__(self, apiKey, apiSecret):
        """Init with Key & Secret."""
        self.apiKey = apiKey
        self.apiSecret = apiSecret

    def __call__(self, r):
        """
        Called when forming a request - generates api key headers. This call uses `expires` instead of nonce.

        This way it will not collide with other processes using the same API Key if requests arrive out of order.
        For more details, see https://www.bitmex.com/app/apiKeys
        """
        # modify and return the request
        expires = int(round(time.time()) + 5)  # 5s grace period in case of clock skew
        r.headers["api-expires"] = str(expires)
        r.headers["api-key"] = self.apiKey
        r.headers["api-signature"] = generate_signature(
            self.apiSecret, r.method, r.url, expires, r.body or ""
        )

        return r


class APIKeyAuth(AuthBase):

    """Attaches API Key Authentication to the given Request object."""

    def __init__(self, apiKey, apiSecret):
        """Init with Key & Secret."""
        self.apiKey = apiKey
        self.apiSecret = apiSecret

    def __call__(self, r):
        """Called when forming a request - generates api key headers."""
        # modify and return the request
        nonce = generate_nonce()
        r.headers["api-nonce"] = str(nonce)
        r.headers["api-key"] = self.apiKey
        r.headers["api-signature"] = generate_signature(
            self.apiSecret, r.method, r.url, nonce, r.body or ""
        )

        return r


def generate_nonce():
    return int(round(time.time() * 1000))


# Generates an API signature.
# A signature is HMAC_SHA256(secret, verb + path + nonce + data), hex encoded.
# Verb must be uppercased, url is relative, nonce must be an increasing 64-bit integer
# and the data, if present, must be JSON without whitespace between keys.
#
# For example, in psuedocode (and in real code below):
#
# verb=POST
# url=/api/v1/order
# nonce=1416993995705
# data={"symbol":"XBTZ14","quantity":1,"price":395.01}
# signature = HEX(HMAC_SHA256(secret, 'POST/api/v1/order1416993995705{"symbol":"XBTZ14","quantity":1,"price":395.01}'))
def generate_signature(secret, verb, url, nonce, data):
    """Generate a request signature compatible with BitMEX."""
    # Parse the url so we can remove the base and extract just the path.
    parsedURL = urlparse(url)
    path = parsedURL.path
    if parsedURL.query:
        path = path + "?" + parsedURL.query

    if isinstance(data, (bytes, bytearray)):
        data = data.decode("utf8")

    # print "Computing HMAC: %s" % verb + path + str(nonce) + data
    message = verb + path + str(nonce) + data

    signature = hmac.new(
        bytes(secret, "utf8"), bytes(message, "utf8"), digestmod=hashlib.sha256
    ).hexdigest()
    return signature
